@@ -4,7 +4,7 @@ import logging
 import symtable
 from collections import defaultdict
 from types import FrameType
-from typing import cast, TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple, Union
+from typing import cast, Any, Dict, List, Optional, Set, Tuple, Union
 
 import astunparse
 import pyccolo as pyc
@@ -81,7 +81,46 @@ blocking_spec = pyc.AugmentationSpec(
 )
 
 
-class SafetyTracer(SingletonBaseTracer):
+class ModuleIniter(pyc.BaseTracer):
+    @pyc.register_raw_handler(pyc.init_module)
+    def init_cell(self, _obj, _node_id, frame: FrameType, *_, **__):
+        nbs().set_name_to_cell_num_mapping(frame)
+        for tracer in pyc._TRACER_STACK:
+            tracer._tracing_enabled_files.add(frame.f_code.co_filename)
+
+
+class StackFrameManager(SingletonBaseTracer):
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.call_depth = 0
+
+    @pyc.register_raw_handler((pyc.call, pyc.return_))
+    def handle_first_ipython_frame(
+        self,
+        _ret: Any,
+        _node_id: None,
+        frame: FrameType,
+        event: pyc.TraceEvent,
+        *_,
+        **__,
+    ):
+        if frame.f_code.co_name == "<traced_lambda>":
+            return pyc.SkipAll
+        # IPython quirk -- every line in outer scope apparently wrapped in lambda
+        # We want to skip the outer 'call' and 'return' for these
+        if event == pyc.call:
+            self.call_depth += 1
+            if self.call_depth == 1:
+                return pyc.SkipAll
+        elif event == pyc.return_:
+            self.call_depth -= 1
+            if nbs().is_develop:
+                assert self.call_depth >= 0
+            if self.call_depth == 0:
+                return pyc.SkipAll
+
+
+class SafetyTracer(StackFrameManager):
     ast_rewriter_cls = SafetyAstRewriter
 
     def should_propagate_handler_exception(
@@ -106,7 +145,6 @@ class SafetyTracer(SingletonBaseTracer):
         self.prev_event: Optional[pyc.TraceEvent] = None
         self.prev_trace_stmt: Optional[TraceStatement] = None
         self.seen_stmts: Set[NodeId] = set()
-        self.call_depth = 0
         self.traced_statements: Dict[NodeId, TraceStatement] = {}
         self.node_id_to_loaded_symbols: Dict[NodeId, List[DataSymbol]] = defaultdict(
             list
@@ -545,11 +583,6 @@ class SafetyTracer(SingletonBaseTracer):
         self.prev_node_id_in_cur_frame = node_id
         self.prev_node_id_in_cur_frame_lexical = node_id
 
-    @pyc.register_raw_handler(pyc.init_module)
-    def init_cell(self, _obj, _node_id, frame: FrameType, _event: pyc.TraceEvent, **__):
-        nbs().set_name_to_cell_num_mapping(frame)
-        self._tracing_enabled_files.add(frame.f_code.co_filename)
-
     # @pyc.register_raw_handler((pyc.before_for_loop_body, pyc.before_while_loop_body))
     # def before_loop_body(self, _obj: Any, loop_id: NodeId, *_, **__):
     #     ret = self.is_tracing_enabled and loop_id not in self._seen_loop_ids
@@ -1226,13 +1259,13 @@ class SafetyTracer(SingletonBaseTracer):
             # position maps to the calling location instead of the return
             return cast(ast.stmt, self.ast_node_by_id[self.next_stmt_node_id])
         try:
-            stmt_node = self.line_to_stmt_by_module_id[cell_num][lineno]
+            stmt_node = self.stmt_by_lineno_by_module_id[cell_num][lineno]
             if event == pyc.call and not isinstance(
                 stmt_node, (ast.AsyncFunctionDef, ast.FunctionDef)
             ):
                 # TODO: this is bad and I should feel bad. Need a better way to figure out which
                 #  stmt is executing than by using line numbers.
-                parent_node = self.parent_node_by_id.get(id(stmt_node), None)
+                parent_node = self.parent_stmt_by_id.get(id(stmt_node), None)
                 if nbs().is_develop:
                     logger.info(
                         "node %s parent %s",
@@ -1256,29 +1289,6 @@ class SafetyTracer(SingletonBaseTracer):
                 raise e
         return None
 
-    @pyc.register_raw_handler((pyc.call, pyc.return_))
-    def handle_first_ipython_frame(
-        self,
-        _ret: Any,
-        _node_id: None,
-        _frame: FrameType,
-        event: pyc.TraceEvent,
-        *_,
-        **__,
-    ):
-        # IPython quirk -- every line in outer scope apparently wrapped in lambda
-        # We want to skip the outer 'call' and 'return' for these
-        if event == pyc.call:
-            self.call_depth += 1
-            if self.call_depth == 1:
-                return pyc.Skip
-        elif event == pyc.return_:
-            self.call_depth -= 1
-            if nbs().is_develop:
-                assert self.call_depth >= 0
-            if self.call_depth == 0:
-                return pyc.Skip
-
     @pyc.register_raw_handler(pyc.call)
     def handle_call(
         self,
