@@ -3,6 +3,7 @@ import ast
 import logging
 from typing import List, Optional, Set, Union
 
+from nbsafety.analysis.live_refs import static_resolve_rvals
 from nbsafety.analysis.symbol_ref import resolve_slice_to_constant
 from nbsafety.analysis.mixins import (
     SaveOffAttributesMixin,
@@ -103,9 +104,9 @@ class ResolveRvalSymbols(
             )
             if dsym is not None:
                 self.symbols.append(dsym)
-        except Exception as e:
-            logger.warning(
-                "Exception occurred while resolving node %s: %s", ast.dump(node), e
+        except Exception:
+            logger.exception(
+                "Exception occurred while resolving node %s", ast.dump(node)
             )
 
     def visit_Subscript(self, node: ast.Subscript):
@@ -144,9 +145,9 @@ class ResolveRvalSymbols(
                     dsym = None
             if dsym is not None:
                 self.symbols.append(dsym)
-        except Exception as e:
-            logger.warning(
-                "Exception occurred while resolving node %s: %s", ast.dump(node), e
+        except Exception:
+            logger.exception(
+                "Exception occurred while resolving node %s", ast.dump(node)
             )
 
     def visit_keyword(self, node: ast.keyword):
@@ -240,6 +241,8 @@ def resolve_rval_symbols(
     if isinstance(node, str):
         node = ast.parse(node).body[0]
     rval_symbols = ResolveRvalSymbols()(node)
+    if len(rval_symbols) == 0:
+        rval_symbols = static_resolve_rvals(node)
     if should_update_usage_info:
         Timestamp.update_usage_info(rval_symbols)
     return rval_symbols
