@@ -36,9 +36,16 @@ def patched_emit_event_fixture():
     ):
         event = evt if isinstance(evt, TraceEvent) else TraceEvent(evt)
         if frame.f_code.co_filename.startswith("<ipython-input"):
+            is_traced_lambda = frame.f_code.co_name == "<traced_lambda>"
             if not (
-                (event == TraceEvent.call and self.call_depth == 0)
-                or (event == TraceEvent.return_ and self.call_depth == 1)
+                (
+                    event == TraceEvent.call
+                    and (self.call_depth == 0 or is_traced_lambda)
+                )
+                or (
+                    event == TraceEvent.return_
+                    and (self.call_depth == 1 or is_traced_lambda)
+                )
             ):
                 if event in self.events_with_registered_handlers:
                     _RECORDED_EVENTS.append(event)
